# # Exploratory script for carbon cycle analysis
#
# ### on the Tegel Project dataset.
# 
# the dataset is represented in the input file 'tp_config.yml'
# 
# #### Naming:
# - corresponding specific files named with prefix 'tp_'
# 
# #### Sidenotes:
# - future to do: should warn if mutliple entries with same material name, as only the first one will be imported 
# (dicts enforce unique keys)
# 
# -------------------------------------

if __name__ == "__main__":

    # ------------------------
    # no need to understand, this is unnesting and converting a yaml file such as the config file

    from typing import Any, Dict, List
    import pandas as pd
    import yaml

    # function and helper-function to dynamicly extract nested dict paths to list of lists
    def _into_rows(nested_dict: Dict[str, Any], path: List[Any], rows: List[List[Any]]):
        for key, value in nested_dict.items():
            innerpath = path + [key]
            if isinstance(value, dict):
                _into_rows(value, innerpath, rows)
            else:
                rows.append(innerpath + [value])

    def _nested_dict_into_rows(nested_dict: Dict[str, Any]) -> List[List[Any]]:
        rows = []
        _into_rows(nested_dict, [], rows)
        return rows

    def yaml_to_dict_of_dfs(nested_dict: Dict[str, Any]) -> Dict[Any, Any]:
        full_out = {}
        for name, sub_config in nested_dict.items():
            if isinstance(sub_config, dict):
                full_out[name] = pd.DataFrame(_nested_dict_into_rows(sub_config))
            else:
                full_out[name] = {name: sub_config}
        return full_out
    # ------------------------

    # -------------------------------------
    # ## Main
    # ### 1 Preconditions
    # packages, input file name and loading data in

    # packages used in code:
    import pandas as pd
    import numpy as np
    import yaml
    # used for graphics:
    import seaborn as sns
    import matplotlib.pyplot as plt

    # display setting for decimals in dataframes:
    pd.set_option('display.float_format', '{:.3f}'.format) # optional

    # configuration file load-in:
    filepath = './tg_config.yml'

    with open(filepath, "r") as file:
        data_dict = yaml.safe_load(file) # contains a list of dicts
        data_items = yaml_to_dict_of_dfs(data_dict) # contains a dict of dataframes
        print(data_items['Name'], 'Dataset', '\nfrom YAML-file:', filepath, '\n')
        print('Data load-in done.')


    # ---------------------------------------------
    # ### 2.1 Initialising the dataframes with column headers
    # 
    # (Column headers are not given in the input yml file)
    # Number of buildings is not given, but calculated by division of total building area by area per building.

    data_items['Material_use'].columns = ['construction_sys', 'building_part', 'material', 'mass_per_building']
    data_items['Material_infos_co2coeff'].columns = ['material','co2coeff']
    data_items['Material_infos_biom_fraction'].columns = ['material', 'biom_fraction']

    # future to do: should warn if mutliple entries with same material name, as only the first one will be imported 
    # (dicts enforce unique keys)

    buildings_total = round(data_dict['area_parameters']['total'] / data_dict['area_parameters']['building']) #167

    # ------------------------
    # ### 2.2 Adding given relevant info to main dataframe
    # 
    # CO2 coefficients and information about the fraction of biomass in given materials are added to the dataframe,
    #  being matched by material names.

    full_df = (data_items['Material_use']
                .merge(data_items['Material_infos_co2coeff'], on='material', how='left')
                .merge(data_items['Material_infos_biom_fraction'], on='material', how='left'))
    print('Initialisation done, yeah!')


    # ---------------------------------------------
    # ### 3 Calculations
    #
    # Adding:
    # - total_masses to full_df (optional, but completes the logic of the dataset)
    # - co2_per_building and adding total_emissions to full_df
    # - c_storage_per_building and adding c_storage_per_building to_full df

    # total masses
    full_df["total_masses"] = (
            full_df['mass_per_building'] # kg / building
            * buildings_total) # number of buildings

    # emissions
    full_df["co2_per_building"] = (
            full_df['mass_per_building'] # kg / building
            * full_df['co2coeff'] # kg CO2equ / kg material
            / 1000) # t CO2equ / kg CO2equ
    full_df["total_emissions"] = (
            buildings_total # number of buildings
            * full_df['co2_per_building']) # kg CO2equ / building

    # storage
    full_df["c_storage_per_building"] = (
            full_df['mass_per_building'] # kg / building
            * full_df['biom_fraction'] # kg biomass / kg material
            * data_dict['Material_infos_carbon_ratio'] # unit C / unit biomass
            / 1000) # t C / kg C
    full_df["total_c_storage"] = (
            buildings_total # number of buildings
            * full_df['c_storage_per_building']) # kg CO2equ / building

    print('Calculations are done.')


    # ---------------------------------------------
    # ### 4 DataFrame Grouping and Export
    # 
    # Grouping the Dataframes to be summed up to construction systems (and building parts)

    cols = ['mass_per_building', 'co2_per_building', 'c_storage_per_building', 'total_masses', 'total_emissions', 'total_c_storage']
    sum_aggr = {f'{col}': "sum" for col in cols}
    df_out = full_df.loc[:,['construction_sys', 'building_part', 'mass_per_building', 'co2_per_building', 'c_storage_per_building', 'total_masses', 'total_emissions', 'total_c_storage']]

    df_out_systems_parts = df_out.groupby(['construction_sys', 'building_part']).agg(sum_aggr).reset_index()
    df_out_systems = df_out.groupby(['construction_sys']).agg(sum_aggr).reset_index()

    # In the following, the main dataframes of interest are being written including masses, emissions and storage:
    # - Values per material. Full Dataset, no sums being aggregated.
    # - Total values per construction systems
    # - Values per construction systems split into the categories of building parts (wall, ceiling, roof)
    #
    # saving to an subfolder 'export'

    print('Writing tables:')
    df_out.to_csv(r'./export/tg_subtotals.csv', sep=',', encoding='utf-8', index=False, header=True, na_rep='NA')
    print(' detailed table to \n "./export/tg_subtotals.csv"')
    df_out_systems.to_csv(r'./export/tg_total_per_csys.csv', sep=',', encoding='utf-8', index=False, header=True, na_rep='NA')
    print(' grouped per construction system to \n "./export/tg_total_per_csys.csv"')
    df_out_systems_parts.to_csv(r'./export/tg_per_csys_and_part.csv', sep=',', encoding='utf-8', index=False, header=True, na_rep='NA')
    print(' grouped per construction system and building part to \n "./export/tg_total_per_csys.csv"')

    # for tab-seperated use sep= '\t'
    # na_rep= can be changed as nedded (e.g. to NA, NaN, 0)



    # ---------------------------------------------
    # ### 5 Graphs and Graph Export
    #
    # This section seeks to visualise the emissions:
    print('Writing plots:')
    # plot
    df_graph1 = (df_out_systems_parts
                .groupby(['construction_sys', 'building_part']).agg({'total_emissions': sum})
                .reset_index()
                .pivot(index='construction_sys', columns='building_part', values='total_emissions'))

    ax = df_graph1.plot(kind= 'bar', stacked=True)

    for c in ax.containers: # value labels
        # Optional: if the segment is small or 0, customize the labels
        labels = ['{0:.0f}kt'.format(v.get_height()/1000) if v.get_height() > 0 else '' for v in c]
        # remove the labels parameter if it's not needed for customized labels
        ax.bar_label(c, labels=labels, label_type='center') #, fmt='%0.2f' could be used but is overwritten by label format statement

    plt.title('Emissions Tegel Project \n167 buildings', fontsize=16) # add overall title
    plt.xlabel('') # add axis titles
    plt.ylabel('material production emissions \nin t CO2')
    plt.xticks(rotation=45) # rotate x-axis labels
    plt.savefig(r'./export/tg_emissions_per_csys_and_part.png', bbox_inches='tight', dpi=300)
    print(' emissions per construction system and building part, stacked \n "./export/tg_emissions_per_csys_and_part.png"')


    # This section seeks to visualise carbon storage:
    #
    # plot
    df_graph2 = (df_out_systems_parts
                .groupby(['construction_sys', 'building_part']).agg({'total_c_storage': sum})
                .reset_index()
                .pivot(index='construction_sys', columns='building_part', values='total_c_storage'))
    df_graph2_filtered = df_graph2[(df_graph2['ceiling'] != 0) & (df_graph2['roof'] != 0) & (df_graph2['wall'] != 0)]

    ax = df_graph2_filtered.plot(kind= 'bar', stacked=True, 
            label= 'plot_storage_per_csys_and_part')

    for c in ax.containers: # value labels
        # Optional: if the segment is small or 0, customize the labels
        labels = ['{0:.0f}kt'.format(v.get_height()/1000) if v.get_height() > 0 else '' for v in c]
        # remove the labels parameter if it's not needed for customized labels
        ax.bar_label(c, labels=labels, label_type='center') #, fmt='%0.2f' could be used but is overwritten by label format statement

    plt.title('Carbon Storage Tegel Project \n167 buildings', fontsize=16) # add overall title
    plt.xlabel('') # add axis titles
    plt.ylabel('carbon storage \nin t C')
    plt.xticks(rotation=45) # rotate x-axis labels
    plt.savefig(r'./export/tg_storage_per_csys_and_part.png', bbox_inches='tight', dpi=300)
    print(' carbon storage per construction system and building part, stacked \n "./export/tg_emissions_per_csys_and_part.png"')

    # ---------------------------------------------
    print('All done.')
# ----------------------------
# The End
