
# Class Approach to CCA
## Specific approach: Tegel project scope

# The following code should fulfill these minimum requirements:
# - handle input as follows:
#     - material info table (.ods)
#     - material usage table (.ods)
#     - material transport information (sheet in material info table)
# - read in data (structured and type-verified in classes)
# - use data to:
#     - built up infrastructures
#     - sum or scale them to project size
#     - create several scenarios with different construction systems
# - create output:
#     - write table file(s)
#     - return df(s)
  

### Notes, Thoughts & To Dos:
# - Object attributes in post init: do they always need to be automaticly calculated? 
#   - Is calling the method when necessary better? 
#   - Is automation inside instantiation here and calling the method there too inconsistent?
# - Infrastructures complementary (& possibly relative?) or alternative for ProjectScope Objects
# - Upscaling of Infrastructures in ProjectScope Objects to actual project parameters dependend on input type of relative weight in material usage, for now only with 'area'
# - some str type args should be changed to Literal['example1', 'example2']


### Dependencies:
# IMPORTANT: following line needs to be changed when standard version includes annotations
from __future__ import annotations # imports future changes that are already in beta, here to have DataClass usable as Type annotation inside itself

# standard library:
from abc import ABC, abstractmethod
from typing import List, Dict, Set, Any, Union, Optional, Literal # type annotations
from math import floor

# non-standard libraries:
import pandas as pd
import numpy as np # pandas uses numpy behind the scenes, whereas we only need it to identify np.nan types
from pydantic.dataclasses import dataclass  # typecheck (obligatory fit, not only hint as from 'dataclasses')
# import yaml # for future use of yaml (again), consider this tool hydra https://hydra.cc/docs/advanced/instantiate_objects/overview/ which already has the support for object instantiation from nested yaml stuctures


### natural (global) variables:
mass_relation_co2_to_c = (2* 15.999 + 12.011) / 12.011 
# atomic masses: carbon = 12; hydrogen = 16
# source: https://www.ciaaw.org/atomic-weights.htm


### Helper Code
# function and helper-function to dynamicly extract nested dict paths to list of lists
# needed if using yaml files in future

# def _into_rows(nested_dict: Dict[str, Any], path: List[Any], rows: List[List[Any]]):
#     for key, value in nested_dict.items():
#         innerpath = path + [key]
#         if isinstance(value, dict):
#             _into_rows(value, innerpath, rows)
#         else:
#             rows.append(innerpath + [value])

# def _nested_dict_into_rows(nested_dict: Dict[str, Any]) -> List[List[Any]]:
#     rows = []
#     _into_rows(nested_dict, [], rows)
#     return rows

# def yaml_to_dict_of_dfs(nested_dict: Dict[str, Any]) -> Dict[Any, Any]:
#     full_out = {}
#     for name, sub_config in nested_dict.items():
#         if isinstance(sub_config, dict):
#             full_out[name] = pd.DataFrame(_nested_dict_into_rows(sub_config)) # converts deep dicts to dataframes
#         else:
#             full_out[name] = sub_config
#     return full_out

### Main Code
# abstract helper class, so that future functions can expect the .to_df method in certain dataclasses
class CCADataClasses(ABC):
    @abstractmethod
    def to_df(self):
        pass

# helper class
@dataclass(frozen=True) # nedded to provide ErrorDimension default values to dataclass attributes. They must be immutable or freshly created via default_factory
class ErrorDimension:
    min: float
    mean: float
    max: float

    def __post_init__(self):
        if not (self.min <= self.mean and self.mean <= self.max):
            raise Exception('Check your min mean max values! Do they follow the logic of 0 <= min <= mean <= max?')
        # elif not self.min >= 0:
        #     raise Exception('Only values >= 0 expected for min')

    def __mul__(self, other) -> ErrorDimension:
        if isinstance(other, ErrorDimension):
            return ErrorDimension(min= self.min * other.min, mean= self.mean * other.mean, max=self.max * other.max)
        else:
            return ErrorDimension(min= self.min * other, mean= self.mean * other, max= self.max * other)

    def __rmul__(self, other) -> ErrorDimension:
        return self.__mul__(other)

    def __truediv__(self, other) -> ErrorDimension:
        if isinstance(other, ErrorDimension):
            return ErrorDimension(min= self.min / other.min, mean= self.mean / other.mean, max=self.max / other.max)
        else:
            return ErrorDimension(min= self.min / other, mean= self.mean / other, max= self.max / other)

    def __rtruediv__(self, other) -> ErrorDimension:
        return self.__mul__(other)

    def __add__(self, other) -> ErrorDimension:
        if isinstance(other, ErrorDimension):
            return ErrorDimension(min= self.min + other.min, mean= self.mean + other.mean, max=self.max + other.max)
        else:
            return ErrorDimension(min= self.min + other, mean= self.mean + other, max= self.max + other)

    def __radd__(self, other) -> ErrorDimension:
        return self.__add__(other)
    
    def __iadd__(self, other) -> ErrorDimension:
        if isinstance(other, ErrorDimension):
            return ErrorDimension(min= self.min + other.min, mean= self.mean + other.mean, max=self.max + other.max)
        else:
            return ErrorDimension(min= self.min + other, mean= self.mean + other, max= self.max + other)

    def __sub__(self, other) -> ErrorDimension:
        if isinstance(other, ErrorDimension):
            return ErrorDimension(min= self.min - other.min, mean= self.mean - other.mean, max=self.max - other.max)
        else:
            return ErrorDimension(min= self.min - other, mean= self.mean - other, max= self.max - other)

    def __rsub__(self, other) -> ErrorDimension:
        return self.__sub__(other)

    def to_df(self, **kwargs) -> pd.DataFrame:
        return pd.DataFrame([[self.min, 'min'], [self.mean, 'mean'],[self.max, 'max']], columns=['value', 'dim'], **kwargs)


# Main classes
@dataclass
class MeanOfTransport(CCADataClasses):
    category: str                               # land, sea, air
    means: str                                  # truck, train, ship, aiplane
    # energy: Optional[float]                     # energy needed to transport the load
    # capacity: Optional[Union[float, Tuple[float, float]]]     # range of or max capacity expected
    co2coeff_transp: ErrorDimension = ErrorDimension(0,0,0) # in tC/tMaterial/km

    def to_df(self):
        return super().to_df()


@dataclass
class MaterialProcessingStep(CCADataClasses): # harvest is also a processing step
    step_in_chain: int                                  # number of processing step in processing chain
    final_material: str                                 # final material
    transport: MeanOfTransport                          # will be globally set to the same mean of transport. Maybe 7.5 t truck as default?
    output_fraction: ErrorDimension = ErrorDimension(1,1,1)                          # = "usecoeff"; fraction of original material weight to be used, e.g. 0.25 eq 25% of material is used
    distance: ErrorDimension = ErrorDimension(0,0,0)                                 # transport distance following the processing step [km] (transporting the output = fraction of input)
    # energy_prod_step: Optional[float]                   # energy needed to process the goods
    co2coef_prod: ErrorDimension = ErrorDimension(0,0,0)     # C emission coefficient  [kg CO2 eq per kg input]
    process_name: Optional[str] = None                  # name of processing step, smth like 'harvest' or 'sawing', 'melting'

    def to_df(self):
        return super().to_df()


@dataclass
class Material(CCADataClasses): # could partly be fetched from a database in future
    name: str
    # specification: Optional[str]
    # category: Optional[str]
    construction_sys: str
    infrastructure_part: str
    weight: float
    processing_chain: Optional[List[MaterialProcessingStep]] = None
    co2coef_prod: Optional[ErrorDimension] = None
    co2coef_transp: Optional[ErrorDimension] = None
    biomass_fraction: ErrorDimension = ErrorDimension(0,0,0)
    ccontent: ErrorDimension = ErrorDimension(0,0,0)

# über field möglichkeit nehmen, das manuell zu geben?
    cstorage: ErrorDimension = ErrorDimension(0,0,0)
    cemissions_prod: ErrorDimension = ErrorDimension(0,0,0)
    cemissions_transp: ErrorDimension = ErrorDimension(0,0,0)
    cemissions_total: ErrorDimension = ErrorDimension(0,0,0)

    def __post_init__(self):
        if self.processing_chain != None:
            self._details_processing_and_transport: Dict[int, Dict[str, ErrorDimension]] = self.process_and_transport() # saving so no need for double computing
        else:
            pass # yet to be implemented
        self.cstorage: ErrorDimension = self.calculate_relative_cstorage()
        self.cemissions_prod: ErrorDimension = self.calc_cemissions(specify= 'cemissions_prod')
        self.cemissions_transp: ErrorDimension = self.calc_cemissions(specify= 'cemissions_transp')
        self.cemissions_total: ErrorDimension = self.cemissions_transp + self.cemissions_prod

    def calculate_relative_cstorage(self) -> ErrorDimension: # 
        return self.biomass_fraction * self.ccontent * self.weight

    def process_and_transport(self) -> Dict[int, Dict[str, ErrorDimension]]:
        out = {} # initialising output
        processing_idx = 0 # initialising counter
        if self.processing_chain:
            for proc_step in self.processing_chain:
                step = {}
                if processing_idx == 0:
                    step["material_output"] = self.weight
                else:
                    step["material_output"] = out[processing_idx -1]["material_input"]
                step["material_input"] = step["material_output"] / proc_step.output_fraction
                # print(step["material_input"],'\n',step['material_input'])
                # print((step["material_input"] - step['material_input']).min >= 0)
                step["waste"] = step["material_input"] - step["material_output"] # this is triggering >= 0 error, for now muted -.-
                step["cemissions_prod"] = (proc_step.co2coef_prod * step["material_input"]) / mass_relation_co2_to_c # from co2coeff to cemissions 
                step["cemissions_transp"] = (proc_step.distance * proc_step.transport.co2coeff_transp * step["material_output"]) / mass_relation_co2_to_c # from co2coeff to cemissions 
                out[processing_idx] = step
                processing_idx += 1
        return out

    def calc_cemissions(self, specify) -> ErrorDimension:
        # mapping
        if specify == 'cemissions_prod':
            # target_attr = self.cemissions_prod
            altern_coef = self.co2coef_prod
        elif specify == 'cemissions_transp':
            # target_attr = self.cemissions_transp
            altern_coef = self.co2coef_transp
        else:
            raise ValueError("specify takes either 'cemissions_prod' or 'cemissions_transp'")
        # calculation
        if self._details_processing_and_transport and not altern_coef:   # processing chain is given (priority)
            target_attr = sum([step[specify] for step in self._details_processing_and_transport.values()], start=ErrorDimension(0,0,0))
        elif altern_coef:           # general cemicoef is given
            target_attr = self.weight * altern_coef / mass_relation_co2_to_c # from co2coeff to cemissions 
        else:                       # not enough given inputs
            raise ValueError("Combination of material inputs invalid. Expected either processing_chain or cemicoef_prod information!")
        # update total cemissions
        self.cemissions_total = self.cemissions_prod + self.cemissions_transp
        return target_attr

    def to_df(self, **kwargs) -> pd.DataFrame:
        single_attr_cols = pd.DataFrame([[self.name, self.construction_sys, self.infrastructure_part, self.weight, 'min'],
                                         [self.name, self.construction_sys, self.infrastructure_part, self.weight, 'mean'],
                                         [self.name, self.construction_sys, self.infrastructure_part, self.weight,'max']],
                                         columns=['material', 'construction_sys', 'part', 'weight', 'dim'])
        df_ed_attrs = [self.cstorage.to_df().rename(columns={'value':'cstorage'}),
                       self.cemissions_prod.to_df().rename(columns={'value':'cemissions_prod'}),
                       self.cemissions_transp.to_df().rename(columns={'value':'cemissions_transp'})]
        df_self = single_attr_cols
        for df in df_ed_attrs:
            df_self = pd.merge(df_self, df, on=['dim'], how='outer')
        return pd.DataFrame(df_self, **kwargs)


@dataclass
class Infrastructure:
    name: str
    construction_sys: str
    materials: List[Material]
    infrastructure_parts: Set[str]
    inhabitants: int = 0
    floorarea: float = 0
    age: int = 0

    cstorage: ErrorDimension = ErrorDimension(0,0,0)
    cemissions_prod: ErrorDimension = ErrorDimension(0,0,0)
    cemissions_transp: ErrorDimension = ErrorDimension(0,0,0)
    cemissions_total: ErrorDimension = ErrorDimension(0,0,0)

    def __post_init__(self):
        self.cstorage: ErrorDimension = self.calculate_cstorage_infra()
        self.cemissions_prod: ErrorDimension = self.calc_cemissions(specify= 'cemissions_prod')
        self.cemissions_transp: ErrorDimension = self.calc_cemissions(specify= 'cemissions_transp')
        self.cemissions_total: ErrorDimension = self.cemissions_transp + self.cemissions_prod

    def calculate_cstorage_infra(self):
        return sum([mat.cstorage for mat in self.materials], start=ErrorDimension(0,0,0))

    def calc_cemissions(self, specify: str) -> ErrorDimension:
        # mapping & calculation
        if specify == 'cemissions_prod' or specify == 'production':
            self.cemissions_prod = sum([mat.cemissions_prod for mat in self.materials], start=ErrorDimension(0,0,0))
            self.cemissions_total = self.cemissions_prod + self.cemissions_transp
            target_attr = self.cemissions_prod
        elif specify == 'cemissions_transp' or specify == 'transport':
            self.cemissions_transp = sum([mat.cemissions_transp for mat in self.materials], start=ErrorDimension(0,0,0))
            self.cemissions_total = self.cemissions_prod + self.cemissions_transp
            target_attr = self.cemissions_transp
        elif specify == 'total' or specify == 'all':
            self.cemissions_prod = sum([mat.cemissions_prod for mat in self.materials], start=ErrorDimension(0,0,0))
            self.cemissions_transp = sum([mat.cemissions_transp for mat in self.materials], start=ErrorDimension(0,0,0))
            self.cemissions_total = self.cemissions_prod + self.cemissions_transp
            target_attr = self.cemissions_total
        else:
            raise ValueError("Expected either \n 'cemissions_prod', 'production', \n or 'cemissions_transp', 'transport', \n or 'total', 'all'")
        # update total cemissions
        return target_attr
    
    def to_df(self, resolution: str = 'infrastructure',**kwargs) -> pd.DataFrame:
        df_infra_cols = pd.DataFrame([[self.name, self.construction_sys, self.inhabitants, self.floorarea, self.age, 'min'],
                                        [self.name, self.construction_sys, self.inhabitants, self.floorarea, self.age, 'mean'],
                                        [self.name, self.construction_sys, self.inhabitants, self.floorarea, self.age, 'max']],
                                        columns=['infrastructure', 'construction_sys', 'inhabitants', 'floorarea', 'age', 'dim'])
        
        if resolution == 'infrastructure':
            df_self = df_infra_cols
            df_ed_attrs = [self.cstorage.to_df().rename(columns={'value':'cstorage'}),
                        self.cemissions_prod.to_df().rename(columns={'value':'cemissions_prod'}),
                        self.cemissions_transp.to_df().rename(columns={'value':'cemissions_transp'})]
            for df in df_ed_attrs:
                df_self = pd.merge(df_self, df, on=['dim'], how='outer')

        elif resolution in ['material', 'materials', 'full']:
            df_self = pd.DataFrame()
            for mat in self.materials:
                df_m_i = pd.merge(df_infra_cols, mat.to_df(), on='dim', how='outer')
                df_self = pd.concat(objs=[df_self, df_m_i], axis= 0, ignore_index= True, join='outer')
        else:
            raise KeyError('Unexpected argument given for "resolution="')

        return pd.DataFrame(df_self, **kwargs)


@dataclass
class ProjectScope:
    name: str
    area: float
    infras_constellation: List[Infrastructure] # Union[List[Infrastructure], Dict[Infrastructure, float]] # future capability for relative input(?): Union[List[Infrastructure], Dict[Infrastructure, float]] 
    scaling: bool = True
    scaling_by: str = 'area'
    inhabitants: int = 0  # future to do: if not provided could refer to a method that sums up infrastructure capita?

    cstorage: ErrorDimension = ErrorDimension(0,0,0)
    cemissions_prod: ErrorDimension = ErrorDimension(0,0,0)
    cemissions_transp: ErrorDimension = ErrorDimension(0,0,0)
    cemissions_total: ErrorDimension = ErrorDimension(0,0,0)

    def __post_init__(self):
        # scaling
        if self.scaling:
            self.cstorage = sum([i.cstorage * self.calc_scaling_factor() for i in self.infras_constellation], start= ErrorDimension(0,0,0))
            self.cemissions_prod = sum([i.cemissions_prod * self.calc_scaling_factor() for i in self.infras_constellation], start= ErrorDimension(0,0,0))
            self.cemissions_transp = sum([i.cemissions_transp * self.calc_scaling_factor() for i in self.infras_constellation], start= ErrorDimension(0,0,0))
            
        else:
            self.cstorage = sum([i.cstorage for i in self.infras_constellation], start= ErrorDimension(0,0,0))
            self.cemissions_prod = sum([i.cemissions_prod for i in self.infras_constellation], start= ErrorDimension(0,0,0))
            self.cemissions_transp = sum([i.cemissions_transp for i in self.infras_constellation], start= ErrorDimension(0,0,0))
        # total cemissions
        self.cemissions_total = self.cemissions_prod + self.cemissions_transp

    def calc_scaling_factor(self): # if scale_by == 'area'
        return floor(self.area / sum(i.floorarea for i in self.infras_constellation)) # math.floor() returns the lower integer (rounds down to no decimal), assuring integrity of infrastructur(sets), as building 0.2 of a house doesn't make sense and and house with factor x of the area would mean material amounts differing from factor x

    def to_df(self, resolution: Literal['projectscope', 'infrastructure', 'material'] = 'projectscope',**kwargs) -> pd.DataFrame:
        df_projectscope = pd.DataFrame([[self.name, self.area, self.inhabitants, 'min'],
                                            [self.name, self.area, self.inhabitants, 'mean'],
                                            [self.name, self.area, self.inhabitants, 'max']],
                                            columns=['project_scope', 'ps_area', 'ps_inhabitants', 'dim'])
        
        if resolution == 'projectscope':
            df_self = df_projectscope
            df_ed_attrs = [self.cstorage.to_df().rename(columns={'value':'cstorage'}),
                        self.cemissions_prod.to_df().rename(columns={'value':'cemissions_prod'}),
                        self.cemissions_transp.to_df().rename(columns={'value':'cemissions_transp'})]
            for df in df_ed_attrs:
                df_self = pd.merge(df_self, df, on=['dim'], how='outer')

        elif resolution == 'infrastructure': 
            scaleby = self.calc_scaling_factor()
            df_self = pd.DataFrame()
            for infra in self.infras_constellation:
                df_infra = infra.to_df()
                for col in ['floorarea', 'inhabitants', 'cstorage', 'cemissions_prod', 'cemissions_transp']:
                    df_infra[col] *= scaleby
                    df_infra.rename(columns={col: f'scaled_{col}'}, inplace=True)
                df_infra_ps = pd.merge(df_projectscope, df_infra, on='dim', how='outer')
                df_self = pd.concat(objs=[df_self, df_infra_ps], axis= 0, ignore_index= True, join='outer')

        elif resolution == 'material':
            scaleby = self.calc_scaling_factor()
            df_self = pd.DataFrame()
            for infra in self.infras_constellation:
                df_infra = infra.to_df(resolution='material')
                for col in ['floorarea', 'inhabitants', 'weight', 'cstorage', 'cemissions_prod', 'cemissions_transp']:
                    df_infra[col] *= scaleby
                    df_infra.rename(columns={col: f'scaled_{col}'}, inplace=True)
                df_infra_ps = pd.merge(df_projectscope, df_infra, on='dim', how='outer')
                df_self = pd.concat(objs=[df_self, df_infra_ps], axis= 0, ignore_index= True, join='outer')

        else:
            raise KeyError('Unexpected argument given for "resolution="')

        return pd.DataFrame(df_self, **kwargs)



### Functions (not methods)

def instantiate_means_of_transport(df:pd.DataFrame) -> List[MeanOfTransport]:
    if 'tCO2' in df.loc[0]: # Expecting a row of units
        df = df.drop([0])
        print("Row with unit details removed while instantiating from transport table input.")
    return [MeanOfTransport(category=x['category'], means=x['transportation'], co2coeff_transp=ErrorDimension(min=x['co2_coeff_min'], mean=x['co2_coeff_mean'], max=x['co2_coeff_max'])) 
            for x in df.drop([0]).to_dict(orient='index').values()]


# this function looks awful wich comes frome it doing all together: validation and instantiation on several levels...
def instantiate_materials(df_use: pd.DataFrame, df_info: pd.DataFrame, mean_of_transport: MeanOfTransport, processing_chain_mode: bool = True) -> List[Material]:
    # input info Dataframe to Dict with names (by logic should be unique!) as keys
    info_dict = {value['material']: value for value in df_info.to_dict(orient='index').values()}
    # usage input data validation: processing steps or general cemicoef ?
    if processing_chain_mode == False:
        if 'cemicoef_prod' and 'cemicoef_transp' not in df_use.columns:
            raise ValueError(f'Input data lacks general carbon emission info in the material usage table {df_use=}.\n Expected headers "cemicoef_prod" and "cemicoef_transp" for {processing_chain_mode=}')
    if processing_chain_mode == True:
        if 'dist1' and 'mass_fraction1' not in df_use.columns:
            raise ValueError(f'Input data lacks processing steps in the material usage table {df_use=}.\n Expected headers "dist1" and "mass_fraction1" for {processing_chain_mode=}')
    # declaration of output
    output_set_of_materials = []
    # material object creation loop
    # with processing steps
    if processing_chain_mode:                                   
        for entry in df_use.to_dict(orient='index').values():   # df to dict, entry = row in input table
            processing_list = []
            restr_proc_dict = {} # restructered
            reduced_dict = {k:v for k,v in entry.items() if 'dist' in k or 'mass_fraction' in k}

            for key, val in reduced_dict.items():
                processing_step = ''.join((filter(str.isdigit, key)))
                pure_key = key.replace(processing_step, '')
                # default value set (not sure if obsolent in future due to default MPS attribute values)
                if np.isnan(val) and 'dist' in pure_key:
                    val = 0
                elif np.isnan(val) and 'mass_fraction' in pure_key:
                    val = 1

                if processing_step not in restr_proc_dict.keys():
                    restr_proc_dict[processing_step] = {}
                restr_proc_dict[processing_step][pure_key] = val

            for pstep, values in restr_proc_dict.items(): # step by step
                processing_list.append(MaterialProcessingStep(final_material= entry['material'],
                                                                step_in_chain= pstep,
                                                                output_fraction= ErrorDimension(min= values['mass_fraction'],
                                                                                         mean= values['mass_fraction'],
                                                                                         max= values['mass_fraction']),
                                                                distance= ErrorDimension(min= values['dist'],
                                                                                         mean= values['dist'],
                                                                                         max= values['dist']),
                                                                transport= mean_of_transport))
            output_set_of_materials.append(Material(name= entry['material'], 
                                                construction_sys= entry['construction_system'],
                                                infrastructure_part= entry['part'],
                                                weight= entry['weight'],
                                                biomass_fraction = ErrorDimension(min= info_dict[entry['material']]['biomass_fraction'],
                                                                                  mean= info_dict[entry['material']]['biomass_fraction'],
                                                                                  max= info_dict[entry['material']]['biomass_fraction']),
                                                ccontent = ErrorDimension(min= info_dict[entry['material']]['carbon_ratio'],
                                                                                  mean= info_dict[entry['material']]['carbon_ratio'],
                                                                                  max= info_dict[entry['material']]['carbon_ratio']),
                                                processing_chain = processing_list,
                                                co2coef_prod= ErrorDimension(min= info_dict[entry['material']]['co2_coeff_min'],
                                                                              mean= info_dict[entry['material']]['co2_coeff_mean'],
                                                                              max= info_dict[entry['material']]['co2_coeff_max'])))
    # with generic carbon emissions coeficients
    else:
        for entry in df_use.to_dict(orient='index').values(): 
            output_set_of_materials.append(Material(name= entry['material'],
                                    construction_sys= entry['construction_system'],
                                    infrastructure_part= entry['part'],
                                    weight= entry['weight'],
                                    co2coef_prod= ErrorDimension(min= entry['cemicoef_prod'],
                                                                 mean= entry['cemicoef_prod'],
                                                                 max= entry['cemicoef_prod']),
                                    co2coef_transp= ErrorDimension(min= entry['cemicoef_transp'],
                                                                   mean= entry['cemicoef_transp'],
                                                                   max= entry['cemicoef_transp']),
                                    biomass_fraction = info_dict[entry['material']]['biomass_fraction'],
                                    ccontent = info_dict[entry['material']]['carbon_ratio']))
    return output_set_of_materials


def filter_materials_by_csys(material_list: List[Material], construction_system: str) -> List[Material]:
    return [mat for mat in material_list if mat.construction_sys == construction_system]


# would be great to combine this with a flexible filter function.
def initialise_infrastructure(material_list: List[Material], construction_system: str, name: str, floorarea: float = 0) -> Infrastructure:
    filtered_list = filter_materials_by_csys(material_list = material_list, construction_system = construction_system)
    return Infrastructure(name= name, 
                            construction_sys= construction_system, 
                            materials= filtered_list, 
                            infrastructure_parts= {mat.infrastructure_part for mat in filtered_list},
                            floorarea= floorarea)


def unique_construction_systems(material_usages: List[Material]) -> List[str]:
    construction_systems_set = set()
    for mat in material_usages:
        construction_systems_set.add(mat.construction_sys)
    return list(construction_systems_set)


def instantiate_infrastructures_by_csys(construction_systems: List[str], full_material_list: List[Material], infra_floorarea: Union[float, Dict[str, float]]) -> List[Infrastructure]:
    if type(infra_floorarea) == float:
        return [Infrastructure(materials= [mat for mat in full_material_list if mat.construction_sys == csys],
                               construction_sys= csys, 
                               infrastructure_parts= {mat.infrastructure_part for mat in [mat for mat in full_material_list if mat.construction_sys == csys]},
                               name= f'infra_{csys}',
                               floorarea = infra_floorarea)       
                for csys in construction_systems]
    elif type(infra_floorarea) == Dict[str, float]:
        if construction_systems in infra_floorarea.keys():
            return [Infrastructure(materials= [mat for mat in full_material_list if mat.construction_sys == csys],
                                   construction_sys= csys,
                                   infrastructure_parts= {mat.infrastructure_part for mat in [mat for mat in full_material_list if mat.construction_sys == csys]},
                                   name= f'infra_{csys}',
                                   floorarea = fa)
                    for csys, fa in infra_floorarea.items()]
        else:
            raise KeyError('The keys do not correspond to the given construction systems.\nExpected dictionary entries like this: "construction_system: flooarea"')
    else:
        raise ValueError('Unexpected Input type for argument "infra_floorarea"')


# function to create multiple project-scope instances at once,
# scaling them in the same moment
def make_construction_sys_scenarios(project_area: float, 
                                    infra_floorarea: Union[float, Dict[str, float]], 
                                    construction_systems: List[str], full_material_list: List[Material], 
                                    infra_list: Optional[List[Infrastructure]]= None, 
                                    print_keys: bool = False
                                    ) -> Dict[str, ProjectScope]:
    scenarios = {}
    if not infra_list:
        infra_list = instantiate_infrastructures_by_csys(construction_systems= construction_systems, full_material_list= full_material_list, infra_floorarea= infra_floorarea)

    for csys in construction_systems:
        scenarios[f'scenario_{csys}'] = ProjectScope(
            name= f'project_scope_{csys}',
            area= project_area,
            infras_constellation= [infra for infra in infra_list if infra.construction_sys == csys])
    if print_keys:
        for k in scenarios.keys():
            print(k)
    return scenarios

# generally useful check if iterables contain the same type
def homogenous_type(seq):
    iseq = iter(seq)
    first_type = type(next(iseq))
    return first_type if all( (type(x) is first_type) for x in iseq ) else False

# use .to_df() methods on lists of objects
def to_df(object_list: list, resolution: Literal['projectscope', 'infrastructure', 'material'] = 'material', seperated: bool = False):
    try:
        if seperated:
            try:
                return [obj.to_df(resolution= resolution) for obj in object_list]
            except:
                raise TypeError(f'One or several items in {object_list} have no method called "to_df"')
        else:
            if homogenous_type(object_list):
                try:
                    df_out = pd.DataFrame()
                    for obj in object_list:
                        df_out = pd.concat(objs=[df_out, obj.to_df(resolution= resolution)], axis= 0, ignore_index= True, join='outer')
                    return df_out
                except:
                    raise ValueError(f'One or several items in {object_list} have no method called "to_df"')
            else:
                raise TypeError(f'{object_list} must contain objects of the same type')
    except:
        raise ValueError()


### Interaction
### Writing and Output
def cca_to_df(
    filepath_minfotable: str,
    filepath_musagetable: str,
    infra_floor_area: float,
    project_area: float,
    write_output: bool = False,
    filename_output: str = "data",  # will be ignored if write_output = False
    resolution: Literal['projectscope', 'infrastructure', 'material'] = 'material'
) -> Union[pd.DataFrame, List[pd.DataFrame]]:
    with open(filepath_minfotable, "rb") as file:
        df_material_info = pd.read_excel(file, "materials").fillna(0)
        df_transport = pd.read_excel(file, "transport").fillna(0)

    with open(filepath_musagetable, "rb") as file:
        df_material_usage = pd.read_excel(file)

    list_means_of_transport = instantiate_means_of_transport(df=df_transport)
    list_material_usage = instantiate_materials(
        df_info=df_material_info,
        df_use=df_material_usage,
        mean_of_transport=list_means_of_transport[0],
    )
    csys_list = unique_construction_systems(material_usages=list_material_usage)

    set_of_scenarios = make_construction_sys_scenarios(
        project_area=project_area,  # not 852277!! (which would be total floor area when 4 floors per building)
        infra_floorarea=infra_floor_area,
        construction_systems=csys_list,
        full_material_list=list_material_usage,
    )

    out = to_df([scenario for scenario in set_of_scenarios.values()], resolution= resolution)

    if write_output is True:
        try:
            if isinstance(out, pd.DataFrame):
                with pd.ExcelWriter(path= f'./{filename_output}.ods') as writer:
                    out.to_excel(writer, index=False, header=True, na_rep='NA')
            elif isinstance(out, list):
                idx = 0
                for obj in out:
                    try:
                        with pd.ExcelWriter(path= f'./{filename_output}_{idx}.ods') as writer:
                            obj.to_excel(writer, index=False, header=True, na_rep='NA')
                    except:
                        raise ValueError(f"Check input types. Couldn't write object number {idx} in {out}: {obj}")
                    idx +=1
        except:
            raise ValueError('Writing output table failed')
    return out


if __name__ == "__main__":
    cca_to_df(
        filepath_minfotable="materialinfo_tegel2.ods",
        filepath_musagetable="materialusage_tegel.ods",
        infra_floor_area=1276.429,
        project_area=213069.25,
        write_output=True
    )

